""" Code provided by Jonathan Masci (jonathan@idsia.ch) 
    
    The code implement the training and sampling of an RBM with binary units
    (visible and hidden).
    It implements both contrastive divergence (CD-k) and persistent contrastive
    divergence (PCD-k)
    
    Permission is granted for anyone to copy, use, modify, or distribute this
    program and accompanying programs and documents for any purpose, provided
    this copyright notice is retained and prominently displayed, along with
    a note saying that the original programs are available from our
    web page.
    The programs and documents are distributed without any warranty, express or
    implied.  As the programs were written for research purposes only, they have
    not been tested to the degree that would be advisable in any important
    application.  All use of these programs is entirely at the user's own risk.
"""

import pycuda.gpuarray as gpuarray
import pycuda.driver as cuda
from pycuda.compiler import SourceModule
import numpy as np
from cublas import sgemm, sgemv
from pycuda.elementwise import ElementwiseKernel
from pycuda.curandom import rand as curand
from pycuda import cumath
import gzip
import cPickle
from plot_utils import tile_raster_images # thanks to the Theano guys
import PIL.Image
import time

# Cuda initialization
cuda.init()

from pycuda.tools import make_default_context
context = make_default_context()
device = context.get_device()

import atexit
atexit.register(context.detach)

# Kernels
sigmoidal = ElementwiseKernel(
        "float *x",
        "x[i] = 1/(1 + exp(-x[i]))",
        "sigmoidal")

sample = ElementwiseKernel(
        "float *x, float*p",
        "x[i] = 1 ? x[i] > p[i] : 0",
        "sample")

class RBM(object):
  def __init__(self, n_vis, n_hid, batch_size, epsilonw=0.1, epsilonvb=0.1, epsilonhb=0.1, weight_cost=0.0002):
    # set parameters
    self.n_vis = n_vis
    self.n_hid = n_hid
    self.batch_size = batch_size
    self.epsilonw = epsilonw
    self.epsilonvb = epsilonvb
    self.epsilonhb = epsilonhb
    self.weight_cost = weight_cost
    
    # init weights
    w_cpu = np.random.uniform(
               low = -4*np.sqrt(6./(n_hid+n_vis)), 
               high = 4*np.sqrt(6./(n_hid+n_vis)), 
               size = (n_vis, n_hid))
    w_cpu = w_cpu.astype(np.float32)
    self.w_gpu = gpuarray.to_gpu(w_cpu)
    
    # init visible units bias
    b_vis_cpu = np.zeros((1, n_vis))
    b_vis_cpu = b_vis_cpu.astype(np.float32)
    self.b_vis_gpu = gpuarray.to_gpu(b_vis_cpu)

    # init hidden units bias    
    b_hid_cpu = np.zeros((1, n_hid))
    b_hid_cpu= b_hid_cpu.astype(np.float32)
    self.b_hid_gpu = gpuarray.to_gpu(b_hid_cpu)
        
    # init storage for weight updates
    self.wu_w = gpuarray.zeros((n_vis, n_hid), dtype=np.float32)
    self.wu_b_vis = gpuarray.zeros((1, n_vis), dtype=np.float32)
    self.wu_b_hid = gpuarray.zeros((1, n_hid), dtype=np.float32)
    self.pos_prods = gpuarray.zeros((n_vis, n_hid), dtype=np.float32)
    self.pos_e_h = gpuarray.zeros((1, n_hid), dtype=np.float32)
    self.pos_e_v = gpuarray.zeros((1, n_vis), dtype=np.float32)
    self.neg_prods = gpuarray.zeros((n_vis, n_hid), dtype=np.float32)
    self.neg_e_h = gpuarray.zeros((1, n_hid), dtype=np.float32)
    self.neg_e_v = gpuarray.zeros((1, n_vis), dtype=np.float32)

    # init temporary storage
    self.v = gpuarray.zeros((batch_size, n_vis), dtype=np.float32)
    self.h = gpuarray.zeros((batch_size, n_hid), dtype=np.float32)

  def prop_up(self):
    # fill the h with the replicated bias values
    # ones(batch_size,1) * b_hid
    sgemm(gpuarray.zeros((self.batch_size, 1), dtype=np.float32) + 1, self.b_hid_gpu, self.h)
    # propup
    sgemm(self.v, self.w_gpu, self.h, beta=1, transa='N', transb='N') # V * W + B_H
    sigmoidal(self.h)
    
  def prop_down(self):
    # fill the v with the replicated bias values
    # ones(batch_size,1) * b_vis
    sgemm(gpuarray.zeros((self.batch_size, 1), dtype=np.float32) + 1, self.b_vis_gpu, self.v)
    # prodown
    sgemm(self.h, self.w_gpu, self.v, transa='N', transb='T', beta=1) # H * W' + BV
    sigmoidal(self.v)
    
  def sample(self, h):
    prob = curand(h.shape)
    sample(h, prob)
    # TODO implement this on cuda
    # h.set(np.random.binomial(1, h.get(), h.shape).astype(np.float32))

  def free_energy(self, v, h, prods, e_h, e_v):
    sgemm(v, h, prods, transa='T', transb='N') # v' * h
    sgemv(h, gpuarray.zeros((h.shape[0],), dtype=np.float32)+1, e_h, trans='T') # sum row
    sgemv(v, gpuarray.zeros((v.shape[0],), dtype=np.float32)+1, e_v, trans='T') # sum row

  def get_error(self, data):
    return gpuarray.sum((data - self.v).__pow__(2)) # gpuarray.sum is over all elements
 
  def weights_update(self):
    self.wu_w = self.epsilonw * ((self.pos_prods - self.neg_prods)/self.batch_size - self.weight_cost * self.w_gpu)
    self.wu_b_vis = (self.epsilonvb / self.batch_size) * (self.pos_e_v - self.neg_e_v)
    self.wu_b_hid = (self.epsilonhb / self.batch_size) * (self.pos_e_h - self.neg_e_h)
    
    self.w_gpu += self.wu_w
    self.b_vis_gpu += self.wu_b_vis
    self.b_hid_gpu += self.wu_b_hid
  
  def gibbs_vhv(self, v):
    self.prop_up()
    self.sample(self.h)
    self.prop_down()

  def gibbs_hvh(self, h, pre_sample, k=1):
    """ Performs a gibbs sampling for k steps starting from the given h sample.
        It stores also in pre_sample the pre sampled activation of the hidden
        units. """
    
    self.h = h
    for t in xrange(k):
      self.prop_down()
      self.sample(self.v)
    
      self.prop_up()
      cuda.memcpy_dtod(pre_sample.gpudata, self.h.gpudata, self.h.mem_size * self.h.dtype.itemsize)
      self.sample(self.h)

  def train(self, n_epochs, data, persistent=False):
    n_batches = data.shape[0]/self.batch_size
    err_sum = 0
    pre_sample = gpuarray.zeros(self.h.shape, dtype=np.float32)

    for epoch in xrange(n_epochs):
      start_time = time.time()
      for batch in xrange(n_batches):
        # get current minibatch on gpu
        self.v = gpuarray.to_gpu(np.array(data[batch*self.batch_size:(batch + 1)*self.batch_size,:]))
        # propup to the hidden (no sampling)
        self.prop_up()
        # stuff for gradient
        self.free_energy(self.v, self.h, self.pos_prods, self.pos_e_h, self.pos_e_v)
        self.sample(self.h)
        
        # if persistent do not reinitialize the chain but use the previous h
        if (persistent != False):
          self.sample(pre_sample)
          self.h = pre_sample
        # gibbs sampling
        self.gibbs_hvh(self.h, pre_sample, 15)
        # stuff for gradient
        self.free_energy(self.v, pre_sample, self.neg_prods, self.neg_e_h, self.neg_e_v)
        
        err = self.get_error(gpuarray.to_gpu(np.array(data[batch*self.batch_size:(batch + 1)*self.batch_size,:])))
        err_sum += err
        # print("Error for batch " + str(batch) + " is " + str(err))
        
        self.weights_update()
      
      print("Error after epoch " + str(epoch) + " is " +  str(err_sum))
      print("Time: " + str(time.time() - start_time))
      
      # Plot filters after each training epoch
      # Construct image from the weight matrix 
      image = PIL.Image.fromarray(tile_raster_images( X = self.w_gpu.get().T,
               img_shape = (28,28),tile_shape = (20,20), 
               tile_spacing=(1,1)))
      image.save('filters_sampled_pcdk/filters_at_epoch_%i.png'%epoch)

    fo = gzip.GzipFile('weights_pcdk.gz', 'wb')
    cPickle.dump([self.w_gpu.get(), self.b_vis_gpu.get(), self.b_hid_gpu.get()], fo)
    fo.close()
    
  def sample_data(self, v, n_samples, each=1000):
    self.v = gpuarray.to_gpu(np.array(v[110:130,:], dtype=np.float32))
    vis_sample = gpuarray.zeros(self.v.shape, dtype=np.float32)
    image = PIL.Image.fromarray(tile_raster_images( X = v[10:30,:],
             img_shape = (28,28),tile_shape = (5,5), 
             tile_spacing=(1,1)))
    image.save('filters_sampled_pcdk/samples_init.png')
    for i in xrange(n_samples):
      for j in xrange(each):
        self.gibbs_vhv(self.v)
        cuda.memcpy_dtod(vis_sample.gpudata, self.v.gpudata, self.v.mem_size * self.v.dtype.itemsize)
        self.sample(self.v)
            
      image = PIL.Image.fromarray(tile_raster_images( X = vis_sample.get(),
               img_shape = (28,28),tile_shape = (5,5), 
               tile_spacing=(1,1)))
      image.save('filters_sampled_pcdk/samples%i.png'%i)      

# Load the dataset 
f = gzip.open('mnist.pkl.gz','rb')
train_set, valid_set, test_set = cPickle.load(f)
f.close()
dat = np.asarray(train_set[0], dtype=np.float32)
dat = dat.astype(np.float32)

layer = RBM(784, 500, 20)

###################################################
# uncomment the following lines to load the weights
###################################################
# fo = gzip.GzipFile('weights_pcdk.gz', 'rb')
# [w, bv, bh] = cPickle.load(fo)
# fo.close()
# w = w.astype(np.float32)
# layer.w_gpu = gpuarray.to_gpu(w)
# bh = bh.astype(np.float32)
# layer.b_hid_gpu = gpuarray.to_gpu(bh)
# bv = bv.astype(np.float32)
# layer.b_vis_gpu = gpuarray.to_gpu(bv)
###################################################

layer.train(15, dat, persistent=True)
layer.sample_data(test_set[0], 10, 1000)
