"""Various utility functions."""

__all__ = ['to_cpu', 'to_gpu', 'sub2ind', 'relative_error', 'psnr']

from ctypes import cast, c_float, c_double, POINTER, Structure
from numpy import complex64
from pycuda.gpuarray import GPUArray, to_gpu

class float2(Structure):
    """ctype wrapper for CUDA float2 datatype."""
    pass
float2._fields_ = [
    ('x', c_float),
    ('y', c_float),
]

class double2(Structure):
    """ctype wrapper for CUDA double2 datatype."""
    pass
double2._fields_ = [
    ('x', c_double),
    ('y', c_double),
]

def pointer(gpuarray):
    """Return the pointer to the linear memory held by a GPUArray object."""
    if gpuarray.dtype != complex64:
        return cast(int(gpuarray.gpudata), POINTER(c_float))
    else:
        return cast(int(gpuarray.gpudata), POINTER(float2))

## Add _as_parameter_ attribute to GPUArray for it to be used as
## arguments to ctypes-wrapped functions.
import pycuda.gpuarray
pycuda.gpuarray.GPUArray._as_parameter_ = property(pointer)

def to_cpu(x):
    """Move x to CPU, if it is on GPU."""
    if isinstance(x, GPUArray):
        return x.get()
    else:
        return x

def to_gpu(x):
    """Move x to GPU, if it is on CPU."""
    if isinstance(x, GPUArray):
        return x
    else:
        return to_gpu(x)

def sub2ind(shape, I, J, row_major=True):
    """Returns the linear indices of subscripts."""
    if row_major:
        ind = (I % shape[0]) * shape[1] + (J % shape[1])
    else:
        ind = (J % shape[1]) * shape[0] + (I % shape[0])
    return ind

def relative_error(result, reference):
    # assert(result.shape == reference.shape)
    from numpy.linalg import norm
    return norm(result - reference) / norm(reference)

def psnr(result, reference):
    from math import log10
    ## Vectorize result and reference
    result    = result.ravel()
    reference = reference.ravel()
    ## Make sure that pixels take values in [0, 255] and not in [0,1]
    assert(1 < max(reference) <= 255)
    mse = sum((result - reference) ** 2) / reference.size
    return 10 * log10(255**2 / mse)
